setfenv(1, ShinyEnv)


ShinyStats = LibStub("AceAddon-3.0"):NewAddon("ShinyStats", "AceConsole-3.0")
_G["ShinyStats"] = ShinyStats

local L = ShinyLocale


function ShinyStats:OnInitialize()
	ShinyDB:OnInitialize()
	ShinyCache:OnInitialize()
	ShinyConfig:OnInitialize()

	self:CreatePatterns()
	self:CreateColors()
	
	self:RegisterChatCommand("shiny", "ChatCommand")
	
	self.tooltipsHooked = false

	self:HookTooltips()
end

	
function ShinyStats:CreateColors()
	local colors = {}
	local color

	for k,data in pairs(ShinyDB.profile.colors) do
		color = Color:CreateFromHEXA(data.value)
		color:SetEnabled(data.enabled)

		colors[k] = color
	end

	self.colors = colors
end

	
function ShinyStats:CreatePatterns()
	local patterns = {
		[SHINY_PATTERN_OTHER] = {},
		[SHINY_PATTERN_EQUIP] = {},
		[SHINY_PATTERN_CHANCE] = {},
		[SHINY_PATTERN_CUSTOM] = {},
		[SHINY_PATTERN_ENCHANT] = {},
		[SHINY_PATTERN_USE] = {},
		[SHINY_PATTERN_STAT] = {},
		[SHINY_PATTERN_PARTIAL] = {},
		[SHINY_PATTERN_SPECIAL] = {},
	}

	for _,data in pairs(PATTERN_MAP) do
		local pattern, id, category = unpack(data)

		if (type(category) == "table") then
			for i,v in ipairs(category) do
				table.insert(patterns[v], data)
			end
		else
			table.insert(patterns[category or SHINY_PATTERN_OTHER], data)
		end
	end

	self.patterns = patterns
end


function ShinyStats:ReloadConfig()
	ShinyCache:Clear()
	self:CreateColors()
end


function ShinyStats:ChatCommand(userInput)
	local userInput = userInput:lower()

	if (userInput == "dbg" or userInput == "debug") then
		ShinyDB.profile.options.DEBUG = not ShinyDB.profile.options.DEBUG
	elseif (userInput == "clearcache" or userInput == "clear") then
		ShinyCache:Clear()
	elseif (userInput == "import") then
		ShinyConfig:ImportConfig()
		ShinyCache:SetTainted(true)
		ShinyStats:CreateColors()
	else
		ShinyConfig:Show()
	end
end


function ShinyStats:IsRecipe(itemLink)
	return (itemLink:match(L["Recipe"]) 
		or itemLink:match(L["Pattern"]) 
		or itemLink:match(L["Plans"]) 
		or itemLink:match(L["Formula"]) 
		or itemLink:match(L["Technique"]) 
		or itemLink:match(L["Design"]) 
		or itemLink:match(L["Manual"]))
end


function ShinyStats:Escape(text)
	return text:gsub("|", "||")
end


function ShinyStats:UpdatePrimaryStat()
	local spec = GetSpecialization();
	local last = self.primaryStat

	if (spec) then
		self.primaryStat = select(7, GetSpecializationInfo(spec))
		-- local primaryStat = select(7, GetSpecializationInfo(spec, nil, nil, nil, UnitSex("player")))
		-- GetSpecializationInfo(GetSpecialization());
	end

	local stat = self.primaryStat
	if (stat ~= last) then
		if (stat and PRIMARY_STAT_MAP[stat]) then
			self.colors[SHINY_PRIMARYSTAT] = self.colors[PRIMARY_STAT_MAP[stat]]
		else
			self.colors[SHINY_PRIMARYSTAT] = nil
		end

		ShinyCache:Clear()
	end
end


function ShinyStats:Strip(text)
	text = text:gsub("|c%x%x%x%x%x%x%x%x", "")
	text = text:gsub("|r", "")

	return text
end


function ShinyStats:Shorten(text)
	return text:gsub(L["%sand%s"], ShinyDB:GetOption("SHORTEN_SEPARATOR"))
end


function ShinyStats:SplitItemLink(itemLink)
	-- item:itemId:enchantId:gemId1:gemId2:gemId3:gemId4:suffixId:uniqueId:linkLevel:specializationID:upgradeId:instanceDifficultyId:numBonusIds:bonusId1:bonusId2:upgradeValue
	-- itemId, enchantId, gemId1, gemId2, gemId3, gemId4, suffixId, uniqueId, linkLevel, specializationID, upgradeId, instanceDifficultyId, numBonusIds, bonusId1, bonusId2, upgradeValue
	return itemLink:match("item:(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*)")
end


function ShinyStats:ValidateSplitItemLink(itemLink)
	-- item:itemId:enchantId:gemId1:gemId2:gemId3:gemId4:suffixId:uniqueId:linkLevel:specializationID:upgradeId:instanceDifficultyId:numBonusIds:bonusId1:bonusId2:upgradeValue
	-- itemId, enchantId, gemId1, gemId2, gemId3, gemId4, suffixId, uniqueId, linkLevel, specializationID, upgradeId, instanceDifficultyId, numBonusIds, bonusId1, bonusId2, upgradeValue
	local result = {}

	for i,v in ipairs{self:SplitItemLink(itemLink)} do
		if (v and v ~= "") then
			result[i] = v
		end
	end
	
	return result[1], result[2], result[3], result[4], result[5], result[6], result[7], result[8], result[9], result[10], result[11], result[12], result[13], result[14], result[15], result[16] 
end


function ShinyStats:ShouldSkip(region)
	local text = region:GetText()
	local r,g,b,a = region:GetTextColor()

	return (text:match(L["^Socket Bonus"]) and r < 0.51 and g < 0.51)				-- Inactive Socket Bonus
		or (text:match(L["^%(*%d*%)*%s*Set"]) and r < 0.51 and g < 0.51)			-- Inactive Set Bonus
		or text:match(L["^Legacy Set"])												-- Legacy Set
		or text:match(L["^<Shift"]) 												-- Shift-Click to ...
		or text:match(L["^<Made"]) 													-- Made by ...
		or text:match(L["^<Right"]) 												-- Right-Click to ...
		or text:match(L["^Season"]) 												-- Season (PvP)
		or text:match(L["^Requires"]) 												-- Requires ...
		or text:match(L["^Durability"]) 											-- Durability
		or text:match(L["^Reforged"])												-- Deprecated (Removed)
end


function ShinyStats:ParseLine(text, category)
	local found, newColor = false

	for _,data in ipairs(self.patterns[category]) do
		local pattern, colorKey = unpack(data)

		-- print("ParseLine ("..category.."):", pattern, text)

		if (text:match(pattern)) then
			found = true
			newColor = colorKey

			-- print("\t", "Match found!")
			break
		end
	end

	return found, newColor
end


function ShinyStats:ParseText(text, ...)		-- categories
	local found, original, newText = false, text
	local replacements = {}

	local i = 0

	for _,category in ipairs{...} do
		for _,data in ipairs(self.patterns[category]) do
			local pattern, colorKey = unpack(data)
			local color = self.colors[colorKey]
			local colorStr

			-- print("ParseText ("..category.."):", pattern, text)

			if (color and color:IsEnabled()) then
				text = text:gsub(pattern, 
					function(match)
						i = i + 1

						-- print("\t", "Match found! ("..i..")")
						replacements[i] = ("|c%s%s|r"):format(color:ToAHEX(), match)

						return "{"..i.."}"
					end)
			end
		end
	end

	if (#replacements > 0) then
		found = true
		newText = text:gsub("{(%d+)}", function(match)
				return replacements[tonumber(match)]
			end)
	end

	return found, newText
end


local lastItem, seen
function ShinyStats:ParseRegion(region, itemLink)
	local options = ShinyDB.profile.options

	if (not (region and region:GetObjectType() == "FontString")) then
		return false
	end

	local original = region:GetText()
	local text = original

	if (not text) then
		return false
	end

	if (ShinyCache:Get(original) == false and ShinyCache:IsEnabled()) then
		-- print("String cached (ignored)")
		return false

	elseif (ShinyCache:Get(original) and ShinyCache:IsEnabled()) then
		-- print("String cached (parsed)")
		local data = ShinyCache:Get(original)
		
		if (data.color and data.color:IsEnabled()) then
			-- print("\tRegion default text color set from cache")
			region:SetTextColor(data.color:ToDECA())
		end

		if (data.text) then
			-- print("\tRegion text set from cache")
			region:SetText(data.text)
		end

		return true
	end

	if (self:ShouldSkip(region)) then
		-- print("Region skipped, matching skip patterns")
		ShinyCache:Set(original, false)
		return false
	end

	local found, newColor, newText = false

	if (text:match(L["^Enchant"])) then
		found, newColor = self:ParseLine(text, SHINY_PATTERN_ENCHANT)

		if (options["SHORTEN_ENCHANT"]) then
			newText = self:Shorten(newText or text)
			-- print("Shortening enchant", text, newText)
		end
		
	elseif (text:match(L["^Equip"])) then
		found, newColor = self:ParseLine(text, SHINY_PATTERN_EQUIP)

		if (options["SHORTEN_EQUIP"]) then
			newText = self:Shorten(newText or text)
		end

	elseif (text:match(L["^Use"])) then
		found, newColor = self:ParseLine(text, SHINY_PATTERN_USE)

		if (options["SHORTEN_USE"]) then
			newText = self:Shorten(newText or text)
		end

	elseif (text:match(L["^Chance"])) then
		found, newColor = self:ParseLine(text, SHINY_PATTERN_CHANCE)

		if (options["SHORTEN_CHANCE"]) then
			newText = self:Shorten(newText or text)
		end

	else
		found, newColor = self:ParseLine(text, SHINY_PATTERN_STAT)

		if (options["SHORTEN_ENCHANT"]) then
			newText = self:Shorten(newText or text)
		end

		if (options["ENABLE_STAT_STRIP_COLORS"]) then
			newText = self:Strip(newText or text)
		end
	end

	if (not found and options["ENABLE_PARTIAL_MATCHES"]) then
		found, newText = self:ParseText(text, SHINY_PATTERN_PARTIAL, SHINY_PATTERN_SPECIAL)

		if (options["SHORTEN_SOCKET_GEM"]) then
			newText = self:Shorten(newText or text)
		end
	end

	-- print("Found:", found)
	local cacheData

	if (found) then
		local color = newColor and self.colors[newColor]
		cacheData = {
			color = color,
			text = newText ~= original and newText
		}

		if (color and color:IsEnabled()) then
			region:SetTextColor(color:ToDECA())
		end

		if (newText ~= original and newText) then
			region:SetText(newText)
		end

		ShinyCache:Set(original, cacheData)

	else
		ShinyCache:Set(original, false)
	end

	-- if (cacheData) then
	-- 	print("CacheData:", cacheData.color, cacheData.text)
	-- end

	if (options["DEBUG"]) then
		if (lastItem ~= itemLink) then
			seen = setmetatable({}, { __mode = "kv" })
			lastItem = itemLink
			print(itemLink)
			seen[region] = true
		elseif (not seen[region]) then
			local pText = newText ~= original and newText and "-->    "..newText or newColor and newColor.id or ""
			print(string.format("%-15s%-45s    %s    %s    %s",found and "(found)" or "", text, pText, ShinyStats:Escape(text), ShinyStats:Escape(pText)))
			seen[region] = true
		end
	end

	return found
end


function ShinyStats:ExtendTooltip(tip, itemLink)
	if (ShinyCache:IsTainted()) then
		ShinyCache:Clear()
	end

	for _,region in ipairs{tip:GetRegions()} do
			self:ParseRegion(region, itemLink)
	end

	tip:Show()
end


function ShinyStats:HookTooltips()
  if (self.tooltipsHooked) then return end
  self.tooltipsHooked = true

  GameTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)
  ItemRefTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)

  ItemRefShoppingTooltip1:HookScript("OnTooltipSetItem", self.TooltipHook)
  ItemRefShoppingTooltip2:HookScript("OnTooltipSetItem", self.TooltipHook)

  ShoppingTooltip1:HookScript("OnTooltipSetItem", self.TooltipHook)
  ShoppingTooltip2:HookScript("OnTooltipSetItem", self.TooltipHook)
  
  if (AtlasLoot and AtlasLoot.db.Tooltip.tooltip == "AtlasLootTooltip" and AtlasLootTooltip) then
  	AtlasLootTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)
  end


  if LinkWrangler and LinkWrangler.RegisterCallback then
    LinkWrangler.RegisterCallback("ShinyStats", function(frame,link)
      ShinyStats:ExtendTooltip(frame,link)
    end, "item")
  end

end


function ShinyStats.TooltipHook(tip)
	local _, itemLink = tip:GetItem()

	if (itemLink) then
		return ShinyStats:ExtendTooltip(tip, itemLink)
	end
end
